# Shebang!
# TIConvertLib v0.3
# By Michael Lee
#
# MASTER PLAN
# > Make this file a module
# > Fill it with every concievable function
# > Create a different file that imports this.
# > In the different file, use optparse to interact with this.
#

import sys
import time
from Axe_Definitions import *


##################
### Exceptions ###
##################
class BufferOverflow(Exception): pass
class EndOfLineError(Exception): pass
class BadData(Exception): pass
class Bad8xp(Exception): pass
class TokensObjectNotInitialized(Exception): pass

############
### Meta ###
############

class MetaInfo:
    """Contains variables with miscellaneous infomation held in strings.

    MetaInfo.axe_version
    MetaInfo.python_version
    MetaInfo.TIConvert_version
    MetaInfo.TIConvert_author
    MetaInfo.TIConvert_release_date
    """
    axe_version = "0.5.0"
    python_version = "3.1"
    TIConvertLib_version = "0.3"
    TIConvertLib_author = "Michael Lee"
    TIConvertLib_release_date = "10 March 2011"
    comment = "Tokenized with TIConvert v" + TIConvertLib_version

##################
### Base Class ###
##################

class Base:
    def __init__(self):
        self.__hexstring = "0123456789ABCDEF"
        return
    
    def _DecToHex(number):
        """Converts a decimal number into a hexadecimal string.
    
        >>> TIConvertLib._DecToHex(12)
        '0C'
        """
        number = number % 65536
        high = number // 256
        low = number % 256
        if high:
            temp = [low // 16,
                    low % 16,
                    high // 16,
                    high %16]
            result = ""
            for i in temp:
                result += self.__hexstring[i]
        if high == 0:
            temp1 = number // 16
            temp2 = number % 16
    
            result = self._hexstring[temp1] + self.__hexstring[temp2]
        return result
    
    def _StrToHex(string):
        """Converts an ascii string into a hex string.
    
        >>> TIConvertLib._StrToHex("Hello World")
        '48656C6C6F20576F726C64'
        """
        result = []
        for char in string:
            result.append(self._DecToHex(ord(char)))
        return "".join(result)
    
    def _HexToDec(hex_):
        temp = str(hex_)
        return self.__hexstring.index(temp[:1]) * 16 + self._hexstring.index(temp[1:])
    
    def _AddPadding(string, total_length, pad_string):
        """Adds padding to the end of a string.
    
        >>> TIConvertLib._AddPadding("Dog", 6, "#")
        'Dog###'
        """
        if len(string) > total_length:
            raise BufferOverflow("AddPadding: String is longer then allocated length")
        for i in range(total_length - len(string)):
            string += pad_string
        return string
    
    def _DoesFileExist(filepath):
        """Checks if a you can read a file.
    
        >>> TIConvertLib._DoesFileExist(r"C:\Calc\thing.txt")
        True
        >>> TIConvertLib._DoesFileExist(r"Z:\Weird\blarg.erv")
        False
        """
        temp = None
        try:
            temp = open(filepath, "rb")
            temp.close()
        except IOError:
            try:
                temp.close()
            except AttributeError:
                pass
            return False    # Exceptions, file does not exist.
        return True         # No exceptions, filepath is valid and does exist.
    
    
############################
### Base Hexstring class ###
############################

class Hexstring(Base, str):
    def __init__(self, input_):
        """Creates a Hexstring object (immutable).

        The only arg must be a string or an int.
        >>> foobar = TIConvertLib.Hexstring("0004080AFF")
        """
        self.__value = None
        if type(input_) == int:
            self.__value = _DecToHex(input_)
        elif type(input_) == str:
            self.__value = input_
            
        self.__linelist = [None]
        for (i, char) in enumerate(self):
            # For line indexing.
            if char == "3F":
                self.__linelist.append(i+1)
        self.__linelist.append(None)
        return
    
    def __repr__(self):
        return "'{0}'".format(self.__value)

    def __str__(self):
        return '{0}'.format(self.__value)

    def __add__(self, other):
        return Hexstring("".join([self, other]))

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.ReturnByte(i, str)
            
    def __len__(self):
        return len(self.__value) // 2
    
    def Slice(self, start=None, finish=None, return_type=None):
        """Returns a substring of the Hexstring object.

        arg1 - Start of slice (optional)
        arg2 - End of slice (optional)
        arg3 - Type of object returned, int, str, or Hexstring.
               Optional, returns Hexstring by default.
        Treat as foobar[:] style slicing.
        >>> foobar = Hexstring("0004080AFF")
        >>> foobar.Slice(1, 2)
        '01'
        >>> foobar.Slice(None, 3)
        '000104'
        >>> foobar.Slice(4,5,int)
        255
        """
        if return_type == int:
            return self.__int_equiv[start:finish]
        else:
            if start != None:
                start = start * 2
            if finish != None:
                finish = finish * 2
            if return_type == str:
                return self.__value[start:finish]
            elif (return_type == Hexstring) or (return_type == None):
                return Hexstring(self.__value[start:finish])
            else:
                raise BadData("Slice() accepts only strings, Hexstrings, or integers as parameters.")
        return

    def NumberOfLines(self):
        return len(self.__linelist)

    def LineSlice(self, line_number, return_type=None):
        return self.Slice(self.__linelist[line_number], self.__linelist[line_number+1], return_type)

    def ReturnByte(self, offset, return_type=int):
        """ Reads a single-byte token and returns a decimal."""
        if return_type == int:
            return self.__int_equiv[offset]
        elif return_type == str:
            return self.Slice(offset, offset+1, str)
        elif return_type == Hexstring:
            return self.Slice(offset, offset+1, Hexstring)
        else:
            raise BadData("ReturnByte accepts only integers, Hexstrings, or strings as parameters.")
        return

    def ReturnTwoBytes(self, offset, return_type=int):
        """Reads a two-byte token and returns a decimal."""
        if return_type == int:
            return self.ReturnByte(offset + 1) * 256 + self.ReturnByte(offset)
        elif return_type == str:
            return self.Slice(offset, offset+2, str)
        elif return_type == Hexstring:
            return self.Slice(offset, offset+2, Hexstring)
        else:
            raise BadData("ReturnTwoBytes accepts only integers, Hexstrings, or strings as paramenters.")
        return

    def ReturnAscii(self, start=None, finish=None):
        temp = self.Slice(start, finish)
        result = []
        for i in range(self.__len__()):
            result.append(chr(temp.ReturnByte(i)))
        return "".join(result)

    def AddPadding(self, new_length, pad_string, return_type=None):
        temp = [self]
        if self.__len__() > new_length:
            temp[0] = self.Slice(0, new_length)
        for i in range(len(temp[0]) - new_length):
            temp.append(pad_string)
        if (return_type == Hexstring) or (return_type == None):
            return Hexstring("".join(temp))
        elif return_type == str:
            return "".join(temp)
        else:
            raise BadData("AddPadding accepts only Hexstrings or strings as parameters.")
        return
    


####################
### Token Object ###
####################
class Tokens(Hexstring):
    """Tokens object - used for manipulating and holding tokens data (8xp)

    Avoid creating or manipulating this object directly - use other functions
    to create this object.  Then manipulate this created object using other functions."""

    ### 8xp File format ###
    # Every 8xp is arranged in this format:
    # In essence, if you were to transfer a program from the calculator to the
    # computer, you would first copy the Actual Data, add the number of tokens
    # to the start (in bytes), wrap it with data (labeled as Preliminary Data
    # in the section below).  This entire part is called the Data Section.
    #
    # Then, the data section is wrapped with more infomation.  The signature and
    # comment are added to the start, along with the length of the Data Section
    # and the checksum is added to the end.
    #
    # The checksum is calculated by adding up the value of every byte in the DATA
    # SECTION ONLY.
    #
    # File
    #     Header
    #         Signature
    #         Comment (added by tokenizers, not important)
    #         Length of Data Section (in bytes)
    #     Data Section
    #         Preliminary info
    #             Unknown parameter (always $0B or $0D)
    #             Length of Variable Data Section (in bytes)
    #             Variable type ID
    #             Variable name (8 chars long, padded with zeros)
    #             Version - usually zero
    #             Length of Variable Data Section (in bytes) (this is a repeat)
    #         Variable Data Section
    #             Number of Tokens
    #             Actual data (finally)
    #     Checksum
    #
    # Sizes:
    # Offset    Length  Description
    # 0         8       **TI83F*
    # 8         3       $1A $0A $00
    # 11        42      Comment added by detokenizer - padded with zeros or spaces.
    # 53        2       Length of data section.  Should be equal to filesize - 57
    #
    # 55        2       Always has a value of $0B or $0D (prefer $0B)
    # 57        2       Length of Variable Data Section
    # 59        1       Variable type ID
    # 60        8       Variable name (padded with zeros)
    # 68        1       Version, usually $00.  (Present if byte @ 55 == $0D)
    # 69        1       Archive flag.  $80 archived, $00 if not. (Present if byte @ 55 == $0D)
    # 70        2       Length of Variable Data Section (repeat of bytes @ 57)
    #
    # 72        2       Number of tokens
    # 74        n       Actual Data
    #
    # 74 + n    2       Checksum
    #######################
    
    def __init__(self):
        # All data is stored as a Hexstring.
        self._full = None
        
        self._signature = Hexstring("2A2A54493833462A1A0A00")      # **TI83F* 0x1A 0x0A 0x00
        temp = "Written using TIConvert v{0}".format(MetaInfo.TIConvertLib_version)
        self._comment = Hexstring(_StrToHex(temp)).AddPadding(42, "00")
        self._data_section_length = None
        
        self._mystery_value = Hexstring("0B")
        self._length_of_variable_data_section = None
        self._variable_type_id = Hexstring("05")
        self._variable_name = None
        self._version = Hexstring("00")
        self._archived_value = Hexstring("00")
        
        self._number_of_tokens = None
        self._actual_data = None
        
        self._checksum = None

        self.__is_initialized = False
        return

    def Open8xp(self, filepath):
        if not(_DoesFileExist(filepath)):
            raise IOError("Filepath does not exist")
        tempfile = open(filepath, "rb")
        tempbuffer = tempfile.read()
        tempfile.close()
        outbuffer = []
        for char in tempbuffer:
            outbuffer.append(_DecToHex(char))
        self._full = Hexstring("".join(outbuffer))

        self._Update(0)
        return

    def ChangeHeaders(self, comment=None,
                            version=None,
                            variable_name=None,
                            variable_type_id = None,
                            number_of_tokens = None):
        def ChangeHeadersTemp(original, new):
            if new == None:
                return original
            else:
                return Hexstring(new)
            return
        ChangeHeadersTemp(self._comment, comment)
        ChangeHeadersTemp(self._version, version)
        ChangeHeadersTemp(self._variable_name, variable_name)
        ChangeHeadersTemp(self._variable_type_id, variable_type_id)
        ChangeHeadersTemp(self._number_of_tokens, number_of_tokens)
        return

    def SetActualData(self, input_):
        if type(input_) == Hexstring:
            self._actual_data = input_
        elif type(input_) == str:
            self._actual_data = Hexstring(input_)
        else:
            raise BadData("Input is not a str or Hexstring")
        return

    def ReturnActualData(self, input_):
        return self._actual_data

    def IsInitialized(self):
        return self._is_initialized
        
    def Slice(self, start=None, finish=None, return_type=Hexstring):
        return self._full.Slice(start, finish, return_type)
        
    def ReturnByte(self, offset, return_type=int):
        """ Reads a single-byte token and returns a decimal."""
        return self._full.ReturnByte(offset, return_type)

    def ReturnTwoBytes(self, offset, return_type=int):
        """Reads a two-byte token and returns a decimal."""
        return self._full.ReturnByte(offset, return_type)

    def _Update(self, type_):
        if type_ == 0:
            # self._full is most recent
            self._signature = self.Slice(0, 11)
            self._comment = self.Slice(11, 53)
            self._data_section_length = self.Slice(53, 55)

            self._mystery_value = self.Slice(55, 57)
            self._length_of_variable_data_section = self.Slice(57, 59)
            self._variable_type_id = self.Slice(59, 60)
            self._variable_name = self.Slice(60, 68)
            self._version = self.Slice(68, 69)
            self._archived_value = self.Slice(69, 70)
            # 70 to 72 is a repeat of self.__length_of_variable_data_section
        
            self._number_of_tokens = self.Slice(72, 74)
            self._actual_data = self.Slice(74, -2)
        
            self._checksum = self.Slice(-2, None)

            self.__is_initialized = True
            return self.__is_initialized     # Everything is done.

        ## Trying to fill in the gaps...
        done = True
        try:
            # Number of Tokens.
            skip = False
            tokens_count = 0
            for char in self._actual_data:
                if skip:
                    skip = False
                else:
                    if type(Axe_Tokens_Main[char]) == dict:
                        skip = True
                        tokens_count += 1
                    else:
                        tokens_count += 1
            tokens_count_hex = Hexstring(_DecToHex(tokens_count))
        except TypeError:
            self.__is_initialized = False

        try:
            # Length of variable data section
            self._length_of_variable_data_section = Hexstring(_DecToHex(len(self._actual_data + 2)))
        except TypeError:
            self.__is_initialized = False

        try:
            temp = self._length_of_variable_data_Section.ReturnTwoBytes(0)
            self._data_section_length = _DecToHex(temp + 19)
        except TypeError:
            self.__is_initialized = False

        try:
            temp = [self._mystery_value,
                    self._length_of_variable_data_section,
                    self._variable_type_id,
                    self._variable_name,
                    self._version,
                    self._archived_value,
                    self._length_of_variable_data_section,
                    self._number_of_tokens,
                    self._actual_data]
            temp = Hexstring("".join())
            checksum = 0
            for byte in temp:
                checksum += Hexstring(byte).ReturnByte(0)
            self._checksum = Hexstring(checksum)
        except TypeError:
            self.__is_initialized = False

        if type_ == 1:
            # self._full needs to be updated.
            temp = [self._signature,
                    self._comment,
                    self._data_section_length,
                    self._mystery_value,
                    self._length_of_variable_data_section,
                    self._variable_type_id,
                    self._variable_name,
                    self._version,
                    self._archived_value,
                    self._length_of_variable_data_section,
                    self._number_of_tokens,
                    self._actual_data,
                    self._checksum]
            try:
                update = "".join(temp)
            except TypeError:
                done = False


        
            self.__is_initialized = True
        return

    def Convert8xpToAxe(self, filepath= None, colons=False, indents=4):
        if filepath != None:
            try:
                temp = open(filepath, "wb")
                temp.close()
            except IOError:
                try:
                    temp.close()
                except:
                    pass
                raise IOError("Filepath does not exist.")
        output = []
        line_list = [[], []]
        indent_list = [0, 0, 0]
        skip = False
        is_first_byte = False
        for i in range(self._actual_data.NumberOfLines() + 1):
            for j in range(indent_list[0]):
                for k in range(indents):
                    output.append(" ")
            output.extend(line_list[0])
                
            line_list[0] = line_list[1]
            line_list[1] = []
            try:
                current_line = self._actual_data.LineSlice(i, Hexstring)
            except IndexError:
                current_line = ""
            indent_list[0] = indent_list[1]
            indent_list[1] = indent_list[2]
            if colons:
                output.append(":")
            for (i, byte) in enumerate(current_line):
                next_byte = current_line.ReturnByte(i+1, Hexstring)
                if i == 0:
                    is_first_byte = True
                if is_first_byte:
                    if byte in Axe_push:
                        # In case of 'If', 'Repeat', 'While:', 'For', etc...
                        indent_list[2] += 1
                    if byte in Axe_pop:
                        # In case of 'End'
                        indent_list[1] -= 1
                        indent_list[2] -= 1
                    if byte in Axe_both_p:
                        if next_byte in Axe_both_p[byte]:
                            # In case of 'Else', 'ElseIf', 'Else!If'
                            indent_list[1] -= 1
                    if byte in Axe_no_p:
                        if next_byte in Axe_no_p[byte]:
                            # In case of 'Return', 'Return!If', 'ReturnIf'
                            pass
                    is_first_byte = False
                          
                if skip:
                    skip = False
                else:
                    if byte in Axe_Tokens_Main:
                        if byte == "3E":
                            is_first_byte = True
                        if type(Axe_Tokens_Main[byte]) == dict:
                            # Two-byte Token
                            skip = True
                            if type(Axe_Tokens_Main[byte][next_byte]) == list:
                                line_list[1].append(Axe_Tokens_Main[byte][next_byte][0])
                            else:
                                line_list[1].append(Axe_Tokens_Main[byte][next_byte])
                        elif type(Axe_Tokens_Main[byte]) == list:
                            line_list[1].append(Axe_Tokens_Main[byte][0])
                        else:                                          
                            line_list[1].append(Axe_Tokens_Main[byte])
                    else:
                        # The token does not exist, apparently.
                        # TODO: add an exception here.
                        pass

        string_result = "".join(output)
            
        if filepath != None:
            temp = open(filepath, "w")
            temp.write(string_result)
            temp.close()
        return string_result

class StringLine(str):
    def __init__(self, input_string):
        self._actual = input_string
        self._newline_index = [None]
        for (i, char) in enumerate(self._actual):
            if char == "\n":
                self._newline_index.append(i+1)
        if self._actual[-1] != "\n":
            self._newline_index.append(None)
        return

    def __str__(self):
        return self._actual

    def __repr__(self):
        return '"{0}"'.format(self._actual)

    def __iter__(self):
        for i in range(len(self._newline_index) - 1):
            yield self._actual[self._newline_index[i]:self._newline_index[i+1]]



def _FlipDict(input_dict):
    """Takes in a dictionary and flips the keys and the values.

    Will handle dictionary in a dictionary, but input dictionary
    must be correctly formatted."""
    new_dict = {}
    max_word_size = 0

    for (token, word) in input_dict.items():
        if type(word) == dict:
            for token2, word2 in input_dict[token].items():
                if type(word2) == list:
                    for i in word2:
                        new_dict[i] = (token + token2)
                        if len(i) > max_word_size:
                            max_word_size = len(i)
                else:
                    new_dict[word2] = (token + token2)
                    if len(word2) > max_word_size:
                        max_word_size = len(word2)
        else:
            if type(word) == list:
                for i in word:
                    new_dict[i] = (token)
                    if len(i) > max_word_size:
                        max_word_size = len(i)
            else:
                new_dict[word] = token
                if len(word) > max_word_size:
                    max_word_size = len(word)

    return (new_dict, max_word_size)

def AxeStringTo8xp(input_string, new_name=None):
    inbuffer = StringLine(input_string)
    
    (Axe_Flipped, max_word_size) = _FlipDict(Axe_Tokens_Main)
    outbuffer = []

    start_single = ['"', '[']
    end_single = ['"', ']']
    string_prefixes = ['"var', '"appv', '"grp', '"pgrm']

    tokens_count = 0
    
    for line in inbuffer:
        if line[-1:] == "\n":
            line = line.strip() + "\n"
        else:
            line = line.strip()
        if line == "...\n":
            outbuffer.append("3A3A3A3F")
            continue
        parse_size = max_word_size
        is_single = False    
        matchup = None
        
        count = 0
        while True:
            char = line[count:count+1]
            if char == "":
                break

            if is_single:
                if char in end_single:
                    if end_single.index(char) == matchup:
                        parse_size = max_word_size
                        matchup = None
                        is_single = False
            else:
                if char in start_single:
                    parse_size = 1
                    matchup = start_single.index(char)
                    is_single = True

            if line[count:count+5] in string_prefixes:
                outbuffer.append(Axe_Flipped['"'])
                outbuffer.append(Axe_Flipped[line[count+1:count+5]])
                count += 5
                tokens_count += 1
                continue
            elif line[count:count+4] in string_prefixes:
                outbuffer.append(Axe_Flipped['"'])
                outbuffer.append(Axe_Flipped[line[count+1:count+4]])
                count += 4
                tokens_count += 1
                continue
                

            while True:
                snip = line[count:count+parse_size]
                if snip in Axe_Flipped:
                    outbuffer.append(Axe_Flipped[snip])
                    count += len(snip)
                    tokens_count += 1
                    if is_single == False:
                        parse_size = max_word_size
                    break
                else:
                    parse_size -= 1


#    result = Tokens()
#    result.SetActualData("".join(outbuffer))
#    result.ChangeHeaders(variable_name=new_name, number_of_tokens=tokens_count)
    
    

#    return (result, tokens_count)
                                     
                    
#test1 = r"C:\Users\Michael\Documents\Programming\Python\TI-decompiler\AMINES3.8xp"
#test2 = r"C:\Users\Michael\Documents\Programming\Python\TI-decompiler\AMINES3_original.txt"
#test3 = r"C:\Users\Michael\Documents\PRogramming\Python\TI-decompiler\AMINES3_new.txt"

#original = Tokens()
#original.Open8xp(test1)
#original_str = original.Convert8xpToAxe()

#temp = open(test2, "w")
#temp.write(original._actual_data + "\n\n" + original_str)
#temp.close()

#(new_tok, i, thing) = AxeStringTo8xp(original_str)
#new = Tokens()
#new.SetActualData(new_tok)
#new_str = new.Convert8xpToAxe()

#temp = open(test3, "w")
#temp.write(new._actual_data + "\n\n" + new_str)
#temp.close()

#print(new._actual_data == original._actual_data)
#print(thing)

#new_test = StringLine(new_str)
#old_test = StringLine(original.Convert8xpToAxe())

#new_buffer = []
#old_buffer = []
#for line in new_test:
#    new_buffer.append(line)
#for line in old_test:
#    old_buffer.append(line)

#for i in range(len(new_buffer)):
#    if new_buffer[i] != old_buffer[i]:
#        print(i)
#        print(new_buffer[i])
#        print(old_buffer[i])

#for i in range(len(new._actual_data)-1):
#    if new._actual_data.ReturnByte(i) != original._actual_data.ReturnByte(i):
#        print(i)
#        print(new._actual_data.ReturnByte(i, str))
#        print(original._actual_data.ReturnByte(i, str))


#stuff




