# Shebang!
# Shebang!
# TIConvert v0.3
# By Michael Lee
# Released 15 March 2011
#

import sys
import optparse
import TIConvertLib

test1 = r"C:\Users\Michael\Documents\Programming\Python\TI-decompiler\Samples\Sample_5.8xp"
test2 = r"C:\Users\Michael\Documents\Programming\Python\TI-decompiler\Samples\Sample_5_out.8xp"

############################
### Command Line Parsing ###
############################
class CommandLine:
    def __init__(self):
        help_string = "TIConvert v0.3, released 15 March 2011.  " \
                      "Written by Michael Lee, using Python v3.1.  " \
                      "Supports converting to Axe v0.5.0.  "
        self.parser = optparse.OptionParser(description = help_string,
                                       add_help_option = True)
        self.parser.add_option("-c", "--colon", "--colons",
                               action = "store_true",
                               default = False,
                               help = "usage: TIConvert.py -c\n" \
                                      "Use this if you want your " \
                                      "text files to start with the colon at " \
                                      "the start of each line.")
        self.parser.add_option("-t", "--tabs", "--tab",
                               action = "callback",
                               callback = self.indent_callback,
                               help = "usage: TIConvert.py -t 4\n" \
                                      "If this is enabled, it indents the code " \
                                      "correctly when converting " \
                                      "from 8xp to txt.  The number " \
                                      "indicates how many spaces each tab " \
                                      "should be.")
        self.parser.add_option("-q", "--quiet",
                               action = "store_true",
                               default = False,
                               help = "usage: TIConvert.py -q\n" \
                                      "If this is enabled, the program " \
                                      "will display absolutely no messages.")
        self.indent_length = 0
        return

    def RunParser(self):
        (options, arg) = self.parser.parse_args()
        self.options = options
        self.arg = arg
        if self.arg == []:
            self.DisplayHelp
        result = {"colons":self.options.colon,
                  "quiet":self.options.quiet,
                  "tabs":self.indent_length}
        return result

    def DisplayHelp(self):
        self.parser.parse_args(["--help"])
        # The above automatically quits after being called.
        return

    def ReturnPaths(self, suffix="_out.txt"):
        input_path = self.arg[0]

        # Find root directory
        cut = input_path.rfind("\\")
        root_path = input_path[0:cut]

        # Find target path
        input_name = input_path[cut:]
        if "." in input_name:
            target_name = input_name[0:input_name.index(".")]
        else:
            target_name = input_name
        target_name += suffix
        target_path = root_path + target_name
        
        return (input_path, target_path)

    def indent_callback(self, option, opt, value, parser):
        # For optparse (handles optional arguments)
        try:
            self.indent_length = int(parser.rargs[0])
            parser.rargs.remove(str(self.indent_length))
        except:
            self.indent_length = 4
        return

#######################
### Other Functions ###
#######################
def write(message, quiet):
    if quiet == False:
        print(message)
    return

#####################
### The Main Code ###
#####################
# Step 1: Reading arguments from the command line.
command_info = CommandLine()
options = command_info.RunParser()
(input_path, target_path) = command_info.ReturnPaths(suffix = "_out.txt")
quiet = options["quiet"]
write("~~~~", quiet)
write("Input directory:  " + input_path + "\n", quiet)
write("Target directory: " + target_path +"\n", quiet)

# Step 2: Preparing to convert the 8xp to Axe.
file = TIConvertLib.Tokens()
file.Open8xp(input_path)

# Step 3: Convert the 8xp to Axe.
write("~~~~", quiet)
write("Starting detokenizing...\n", quiet)
file.Convert8xpToAxe(target_path, options["colons"], options["tabs"])
write("Finished detokenizing...\n", quiet)

# Step 4: Clean up.
write("~~~~", quiet)
write("Successful and done.\n", quiet)

sys.exit(0)




